from django.apps import AppConfig
from django.utils.module_loading import autodiscover_modules

class TransactionalMailConfig(AppConfig):
    name = 'transactional_mail'

    def ready(self):
        # Automatically discover any emails in installed apps
        autodiscover_modules('emails')


from bs4 import BeautifulSoup
from django.contrib import admin
from django.urls import path
from django.template.response import TemplateResponse
from django.template.loader import render_to_string
from django.http.response import HttpResponse, Http404

from .models import SentMail
from . import registry
from .registry import _default_registry

@admin.register(SentMail)
class SentMailAdmin(admin.ModelAdmin):
    list_display = ('subject', 'email_address', 'user', 'timestamp')



class RegisteredEmailType(SentMail):
    class Meta:
        proxy = True


@admin.register(RegisteredEmailType)
class RegisteredEmailAdmin(admin.ModelAdmin):
    def changelist_view(self, request, **kwargs):
        response = super().changelist_view(request, **kwargs)
        response.context_data['summary'] = _default_registry._registry.keys()

        return response

    def get_urls(self):
        urls = [
            path('<slug:name>/preview/',
                self.admin_site.admin_view(self.preview),
                name="transactional_mail_preview"
            )
        ] + super().get_urls()
        return urls

    def preview(self, request, name):
        klass = registry.get_handler(name)
        if not klass:
            raise Http404()

        email = klass.get_for_preview(request)

        if request.GET.get('plain'):
            text = "Subject: %s\n\n%s" % (email.subject, email.plain)
            return HttpResponse(text, content_type="text/plain")
        else:
            # Insert an extra table with email metadata
            extra = render_to_string(
                'admin/transactional_mail/preview_table.html',
                {'email': email}
            )
            soup = BeautifulSoup(email.html, 'html.parser')
            if soup.head.title:
                title = soup.head.title
            else:
                title = soup.new_tag('title')
                soup.head.insert(0, title)
            title.string = email.subject

            soup.body.insert(0, BeautifulSoup(extra, 'html.parser'))
            return HttpResponse(soup.encode())



"""Models related to sending emails"""

from django.contrib.auth import get_user_model
from django.db import models

class SentMail(models.Model):
    """A log of sent emails"""
    class_name = models.CharField(max_length=64)
    email_address = models.EmailField()
    user = models.ForeignKey(get_user_model(), on_delete=models.SET_NULL, null=True, blank=True)
    subject = models.CharField(max_length=255)
    plain_body = models.TextField()
    html_body = models.TextField()
    timestamp = models.DateTimeField(auto_now_add=True)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='django-transactional-mail',
    version='0.2.2',
    author='Vinay Pai',
    author_email='vinay@vinaypai.com',
    description='Transactional emails in Django projects made easy',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/vinaypai/django-transactional-mail/',
    packages=['transactional_mail'],
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Communications :: Email',
    ],
    install_requires=[
        'django >= 2.2',
        'django-render-block >= 0.5',
        'html2text >= 2017',
        'cssutils >= 1.0.2'
    ]
)



"""Base classes for transactional emails"""
from urllib.parse import urljoin

from django.conf import settings
from django.core.mail import mail_admins
from django.core.mail import send_mail
from django.db.utils import ProgrammingError
from html2text import HTML2Text
from render_block import BlockNotFound, render_block_to_string
from xml.etree import ElementTree

from .models import SentMail

class Email:
    template = None
    from_email = settings.DEFAULT_FROM_EMAIL
    user = None

    def __init__(self, ctx, to_email):
        if self.template is None:
            raise ProgrammingError("Email subclasses must set template")

        self.to_email = to_email

        ctx = {'BASE_URL': settings.BASE_URL, **ctx}
        self.ctx = ctx

        self.subject = render_block_to_string(self.template, 'subject', ctx).strip()

        # Make links absolute
        self.html = render_block_to_string(self.template, 'html', ctx)
        self._make_links_absolute()

        try:
            self.plain = render_block_to_string(self.template, 'plain', ctx).strip()
        except BlockNotFound:
            h = HTML2Text()
            h.ignore_images = True
            h.ignore_emphasis = True
            h.ignore_tables = True
            self.plain = h.handle(self.html)

    def send(self):
        send_mail(
            self.subject,
            self.plain,
            self.from_email,
            [self.to_email],
            html_message=self.html
        )

        log = SentMail(
            class_name=self.__class__.__name__,
            email_address=self.to_email,
            subject=self.subject,
            plain_body=self.plain,
            html_body=self.html
        )
        if self.user:
            log.user = self.user
        log.save()

        self.after_send(log=log) # pylint: disable=no-member

    def after_send(self, *args, **kwargs):
        """Called after an email is sent"""

    @classmethod
    def get_for_preview(cls, request):
        return cls({}, 'nobody@example.com')

    def _make_links_absolute(self):
        """Make all src and href attrs in the html document absolute"""
        root = ElementTree.fromstring(self.html)

        for att in ('src', 'href'):
            for tag in root.findall(f'.//*[@{att}]'):
                tag.attrib[att] = urljoin(settings.BASE_URL, tag.attrib[att])

        self.html = ElementTree.tostring(root, method="html", encoding="unicode")

class UserEmail(Email):
    def __init__(self, ctx, user):
        self.user = user
        ctx = {'user': user, **ctx}

        super().__init__(ctx, user.email)


class AdminEmail(Email):
    def __init__(self, ctx):
        super().__init__(ctx, None)

    def send(self):
        mail_admins(
            self.subject,
            self.plain,
            html_message=self.html
        )



"""
Implements a registry for previews similar to the ModelAdmin registry in
the standard Django Admin app.
"""

from .emails import Email

class AlreadyRegistered(Exception):
    "Raised if a handler is registered to a previously registered owner"

class Registry:
    """
    The JobResultAnalyzer registry is analogous to the admin site registry.
    Register handlers to owners. Handlers should be subclasses of
    JobResultAnalyzer.
    """
    def __init__(self):
        self._registry = {}

    def register(self, name, klass):
        """Register an email class for preview"""
        if not issubclass(klass, Email):
            raise ValueError("Wrapped class must subclass Email")

        if name in self._registry:
            raise AlreadyRegistered()

        self._registry[name] = klass

    def get_handler(self, name):
        """Get the class for this name"""
        return self._registry.get(name)


_default_registry = Registry() #pylint: disable=invalid-name

def register(name):
    """Register the given object the wrapped handler class"""
    def wrapper(email_class):
        _default_registry.register(name, email_class)

        return email_class
    return wrapper


def get_handler(name):
    """Get the handler class for this name from the default registry"""
    return _default_registry.get_handler(name)
